#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <errno.h>

#include "preview.h"

/* GOALS: */
unsigned xres;      /* pixels */
unsigned yres;      /* pixels */
unsigned refresh;   /* Hz */

/* PARAMETERS: */
unsigned hsyncfreq; /* 30..70 kHz  (as Hz) */
unsigned dispclock; /* 25..155 MHz (as Hz) */

unsigned left=0,right=0, top=0,bottom=0;
static double vslength = 46e-6; /* vertical sync: 46 microseconds */
static double hslength =  1e-6; /* horizontal sync: 1 microsecond */

/* RESULTS: */
unsigned hslen, vslen;

/* LIMITS: */
static unsigned mclk = 83e6;       /* atyfb vars, ? */
static unsigned xtal = 14.31818e6; /* atyfb vars, ? */
static unsigned pll  = 83e6;       /* atyfb vars, ? */

static unsigned hsyncmin = 30e3; /* Minimum hsync freq (Hz) */
static unsigned hsyncmax = 70e3; /* Maximum hsync freq (Hz) */
static unsigned vsyncmin = 50;   /* Minimum refresh rate (Hz) */
static unsigned vsyncmax = 120;  /* Maximum refresh rate (Hz) */

static char *ArgLine = NULL;

static void GuessClocks(unsigned xblank, unsigned yblank)
{
    /* We want about 47 microseconds for the vertical sync.
     *
     * Length of (yres + yblank + yslen) = 1/refresh seconds.
     * 
     * Length of yslen is yslen*(1/refresh)/(yres+yblank+yslen)
     *
     * yslen*(1/refresh)/(yres+yblank+yslen) = 47e-6
     *
     * yslen = (yres+yblank) * 47e-6 / (1/refresh - 47e-6)
     */
    
    vslen = (unsigned) ((yres+yblank) * vslength / (1.0/refresh - vslength));
    
    /* hsyncfreq = (yslen + yblank + yres) * refreshfreq
     */
    
    hsyncfreq = (vslen + yblank + yres) * refresh;
    
    /* We want about 1 microseconds for the horizontal sync.
     * 
     * Length of (xres + xblank + xslen) = 1/hsyncfreq seconds.
     *
     * xslen = (xres+xblank) * 1e-6 / (1/hsyncfreq - 1e-6)
     */
    
    hslen = (unsigned) ((xres+xblank) * hslength / (1.0/hsyncfreq - hslength));
    
    dispclock = refresh * (xres+xblank+hslen) * (yres+yblank+vslen);
}

static void VerifyMode(void)
{
    unsigned pixrate;
    unsigned ref_clk_per, pll_per, mclk_per, vclk_per, q;
    
    fflush(stdout);
    if(refresh < vsyncmin) { fprintf(stderr, "# Too low vsync (%u < %u).\n", refresh, vsyncmin); }
    if(refresh > vsyncmax) { fprintf(stderr, "# Too high vsync (%u > %u).\n", refresh, vsyncmax); }
    pixrate = xres * yres * refresh;
    if(hsyncfreq < hsyncmin)
    {
        fprintf(stderr, "# Too low hsync (%u < %u)", hsyncfreq, hsyncmin);
        fprintf(stderr,
            ". Simplest choices:\n");
        
        fprintf(stderr,
            "#   Increase vsync (-r) to at least %u\n",
            hsyncmin / (vslen + (top+bottom) + yres));
        fprintf(stderr,
            "#   Increase blankpart (-y) to at least %u\n",
                hsyncmin/refresh - yres - vslen);
        
        fprintf(stderr, "\n");
    }
    if(hsyncfreq > hsyncmax)
    {
        fprintf(stderr, "# Too high hsync (%u > %u)", hsyncfreq, hsyncmax);
        fprintf(stderr,
            ". Simplest choices:\n");
        
        fprintf(stderr,
            "#   Decrease vsync (-r) to at most %u\n",
            hsyncmax / (vslen + (top+bottom) + yres));
        fprintf(stderr,
            "#   Decrease blankpart (-x) to at most %u\n",
                hsyncmax/refresh - yres - vslen);
        
        fprintf(stderr, "\n");
    }
    
    if(1.0/hsyncfreq < hslength) { fprintf(stderr, "IMPOSSIBLE MODE: scanline is shorter (%.4fs) than hsync time (%.1f s)\n", 1e6/hsyncfreq, hslength*1e6); }
    if(1.0/refresh  < vslength) { fprintf(stderr, "IMPOSSIBLE MODE: screen update is shorter (%.4fs) than vsync time (%.1f s)\n", 1e6/refresh, vslength*1e6); }
    
    /* These errors only apply for atyfb */
    if(xres/8 - 1 > 0xFF)            { fprintf(stderr, "# Mode invalid for ATYFB: xres (%u) > 2048 (h_disp too large)\n", xres); }
    if(xres/8 - 1 + right/8 > 0x1FF) { fprintf(stderr, "# Mode invalid for ATYFB: xres+right (%u) > 4096 (h_sync_start too large)\n", xres+right); }
    if((hslen+7) / 8 > 0x1F)         { fprintf(stderr, "# Mode invalid for ATYFB: hslen (%u) > 248 (h_sync_wid too large)\n", hslen); }

    if(yres - 1 > 0x7FF)             { fprintf(stderr, "# Mode invalid for ATYFB: yres (%u) > 2048 (v_disp too large)\n", xres); }
    if(yres - 1 + bottom > 0x7FF)    { fprintf(stderr, "# Mode invalid for ATYFB: yres+bottom (%u) > 2048 (v_sync_strt too large)\n", yres+bottom); }
    if(vslen > 0x1F)                 { fprintf(stderr, "# Mode invalid for ATYFB: vslen (%u) > 31 (v_sync_wid too large)\n", vslen); }
    
    ref_clk_per = 1e12 / xtal;
    pll_per     = 1e12 / pll;
    mclk_per    = 1e12 / mclk;
    vclk_per    = 1e12 / dispclock;
    
    q = pll_per * 255 / vclk_per;
    if(q < 16)                       { fprintf(stderr, "# Mode invalid for ATYFB: vclk_per too small.\n"); }
    if(q > 255)                      { fprintf(stderr, "# Mode invalid for ATYFB: dispclock too small.\n"); }
}

static void DumpMode_fbset(void)
{
    fflush(stdout);
    fprintf(stderr, "# fbset mode:\n");
    if(ArgLine)
        fprintf(stderr, "# Generated with: modetest %s\n", ArgLine);
    fprintf(stderr,
        "mode \"%ux%u-%u\"\n"
        "    # D: %.3f MHz, H: %.3f kHz, V: %.3f Hz\n"
        "    geometry %u %u   %u %u   %u\n"
        "    timings %u  %u %u  %u %u  %u %u\n"
        "    vsync high\n"
        "    hsync high\n"
        "endmode\n",
        xres, yres, refresh,
        dispclock/1e6,
        hsyncfreq/1e3,
        refresh/1.0,
        xres, yres,  xres, yres,  8,
        (unsigned)(1e12 / dispclock),
        left, right,
        top, bottom,
        hslen, vslen);
}

static void DumpMode_xf86(void)
{
    fprintf(stderr,
        "# XF86Config modeline\n"
        "Modeline \"%ux%u-%u\"     %-8.3f %u %u %u %u   %u %u %u %u\n",
        xres, yres, refresh,
        dispclock / 1e6,
        xres, left+xres, left+xres+hslen, left+xres+hslen+right,
        yres, top+yres,  top+yres+vslen,  top+yres+vslen+bottom);
}

static void DumpMode_stm_sub(unsigned xw, unsigned yw)
{
    fprintf(stderr,
        "\"%ux%u-%ux%u-%uHz\"     %-8.3f %u %u %u %u   %u %u %u %u  font %ux%u\n",
        xres/xw, yres/yw, xw,yw, refresh,
        hsyncfreq/1e3,
        xres, left+xres, left+xres+hslen, left+xres+hslen+right,
        yres, top+yres,  top+yres+vslen,  top+yres+vslen+bottom,
        xw, yw);
}
static void DumpMode_stm(void)
{
    fprintf(stderr, "# TextConfig modelines\n");

    DumpMode_stm_sub(8, 8); DumpMode_stm_sub(9, 8);
    DumpMode_stm_sub(8,14); DumpMode_stm_sub(9,14);
    DumpMode_stm_sub(8,16); DumpMode_stm_sub(9,16);
}

#include <argh.h>

static void PrintVersionInfo(void)
{
    printf("modetest - video mode composer v"VERSION" (C) 1992,2002 Bisqwit\n");
}

static void ArghValueError(long l)
{
    fprintf(stderr, "modetest: parameter value out of range: `%ld'\n", l);
}

static void ArghValueErrorS(const char *s)
{
    fprintf(stderr, "modetest: invalid value: `%s'\n", s);
}


/* This piece of code holds the stderr output until the program is finished. */
#include <unistd.h>
#include <sys/wait.h>
static void PipeEnd(void)
{
    fclose(stderr);
    wait(NULL);
}
static void PipeError(void)
{
    int pip[2], pid;
    unsigned buflen;
    char *buf;
    if(pipe(pip) < 0)return;
    pid = fork();
    if(pid < 0) { close(pip[0]); close(pip[1]); return; }
    if(pid)
    {
        fflush(stderr);
        dup2(pip[1], STDERR_FILENO); /* Reroute stderr to this pipe */
        close(pip[0]); close(pip[1]);
        atexit(PipeEnd);
        return;
    }
    /* The child still has the original stderr. */
    close(pip[1]); /* Close the one we're not going to read */
    for(buflen=0, buf=NULL; ; )
    {
        char Buf[4096], *newbuf;
        int got = read(pip[0], Buf, sizeof Buf);
        if(got <= 0)break; /* eof or error */
        newbuf = realloc(buf, buflen + got);
        if(!newbuf)break; /* alloc failed */
        memcpy((buf = newbuf) + buflen, Buf, (unsigned) got);
        buflen += got;
    }
    fwrite(buf, buflen, 1, stderr);
    if(buf)free(buf);
    _exit(0);
}
/* End of stderr buffer code. */

static void ArgInit(int argc, const char *const *argv)
{
    unsigned len = 0;
    int a;
    char *args;
    for(a=1; a<argc; ++a)len += 1+strlen(argv[a]);
    args = (char *)malloc(len);
    *args = 0;
    for(a=1; a<argc; ++a)
    {
        static const char Buf[2] = " ";
        if(a > 1)strcat(args, Buf);
        strcat(args, argv[a]);
    }
    ArgLine = args;
}

int main(int argc, const char *const *argv)
{
    int Heelp = 0;
    
    if(isatty(STDERR_FILENO))
    {
        /* stderr is console.
         * Therefore we start a child process to
         * buffer all the output and write it out
         * not before the program exits.
         */
        PipeError();
    }
    
    ArgInit(argc, argv);
    
    argh_init();
    
    argh_add_long("help",   'h'); argh_add_bool('h'); argh_add_desc('h', "Help", NULL);
    argh_add_long("version",'V'); argh_add_bool('V'); argh_add_desc('V', "Version information", NULL);
    argh_add_long("xres",   'x'); argh_add_string('x', 1, 64); argh_add_desc('x', "Horizontal resolution.", "<visible>[+<invisible>[/<left%>]]");
    argh_add_long("yres",   'y'); argh_add_string('y', 1, 64); argh_add_desc('y', "Vertical resolution.", "<visible>[+<invisible>[/<top%>]]");
    argh_add_long("refresh",'r'); argh_add_int('r',1,65535); argh_add_desc('r', "Screen refresh rate.", "<num>");
    argh_add_long("hslen",  'X'); argh_add_float('X', 0.0, 1.0); argh_add_desc('X', "Horizontal sync length. Range: 0..1 s, default: 1e-6 (1 microsecond)", "<num>");
    argh_add_long("vslen",  'Y'); argh_add_float('Y', 0.0, 1.0); argh_add_desc('Y', "Vertical sync length. Range: 0..1 s, default: 46e-6 (46 microseconds)", "<num>");
    
    argh_start_parse(argc, argv);
    
    for(;;)
    {
        int c = argh_get_param();
        if(c == -1)break;
        switch(c)
        {
            case 'V':
                PrintVersionInfo();
                return 0;
            case 'h':
                Heelp = 1;
                break;
            case 'x':
            {
                const char *s = argh_get_string();
                unsigned blank;
                char *ep;
                long l = strtol(s, &ep, 10);
                if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                xres = l;
                if(*ep == '+')
                {
                    double perc = 80.0;
                    l = strtol(s=ep+1, &ep, 10);
                    if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                    blank = l;
                    if(*ep == '/')
                    {
                        errno = 0;
                        perc = strtod(ep+1, &ep);
                        if((ep && *ep) || perc<0 || perc>100.0) { ArghValueErrorS(ep); return -1; }
                    }
                    left = blank*perc/100.0;
                    right = blank-left;
                }
                if(*ep) { fprintf(stderr, "modetest: Parse error before `%s'\n", ep); return -1; }
                break;
            }
            case 'y':
            {
                const char *s = argh_get_string();
                unsigned blank;
                char *ep;
                long l = strtol(s, &ep, 10);
                if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                yres = l;
                if(*ep == '+')
                {
                    double perc = 90.0;
                    l = strtol(s=ep+1, &ep, 10);
                    if(l < 1 || l > 32767) { ArghValueError(l); return -1; }
                    blank = l;
                    if(*ep == '/')
                    {
                        errno = 0;
                        perc = strtod(ep+1, &ep);
                        if((ep && *ep) || perc<0 || perc>100.0) { ArghValueErrorS(ep); return -1; }
                    }
                    top = blank*perc/100.0;
                    bottom = blank-top;
                }
                if(*ep) { fprintf(stderr, "modetest: Parse error before `%s'\n", ep); return -1; }
                break;
            }
            case 'X':
                hslength = argh_get_float();
                break;
            case 'Y':
                vslength = argh_get_float();
                break;
            case 'r':
                refresh = argh_get_int();
                break;
            default:
                fprintf(stderr,
                    "modetest: Parameter `%s' is unrecognized.\n",
                    argh_get_string());
                return -1;
        }
    }
    if(!argh_ok())return -1;
    
    if(!xres || !yres || !refresh)
    {
        fprintf(stderr, "Missing options.\n");
        Heelp=1;
    }
    
    if(Heelp)
    {
        PrintVersionInfo();
        printf(
            "\nDevelopes a screen mode."
            "\n");
        printf(
            "Usage: modetest <options> [<...>]\n"
            "Options:\n");
        argh_list_options();
        printf(
            "\n"
            "You must specify at least xres, yres and refresh.\n"
            "Current assumed monitor and display card limits (used in warnings):\n"
            "    Horizontal sync (hsync) range: %u .. %u kHz\n"
            "    Vertical sync (refresh) range: %u .. %u Hz\n"
            "    atyfb maxs: h_disp=2048, h_sync_start=4096, hslen=248,\n"
            "                v_disp=2048, v_sync_strt=2048, vslen=31\n",
                hsyncmin/1000, hsyncmax/1000,
                vsyncmin, vsyncmax
        );
        printf(
            "\n"
            "The program tries to set the screenmode with fbset command.\n"
            "You might want to redirect the stderr to a file in case the fbset command\n"
            "fails and renders the screen garbled: modetest [<options>] 2>logfile.txt\n"
            "\n"
            "Example: modetest 2>tmptmp -x1280+296/78.4 -y1024+39/99 -r60\n"
            "         (gives the screenmode I'm using with my framebuffer console)\n"
        );
        return 0;
    }
    
    GuessClocks(left+right, top+bottom);
    VerifyMode();
    DumpMode_fbset();
    DumpMode_xf86();
    DumpMode_stm();
    
    FillScreen();
    SetMode();
    FillScreen();
    
    getchar();
    
    RestoreMode();
    
    return 0;
}
